#!/usr/bin/python

"""
A program that opens a TCP connection to port 4606 on a DeviceMaster
and uses admin commands to download a binary into RAM on the
DeviceMaster (and optionally run it).
"""

usageString = \
"""
 Usage: %s [-gqsr] [-b <base>] [-e <start>] hostname filename
            -g   send go command to start program
            -q   quiet mode (do not show packets)
            -r   reset DM before downloading
            -s   skip vector area (use with gdb)
            -b N base address to which the file should be loaded
            -e N entry point to jump to instead of base address

 Example:
            download -rg 10.0.0.108 nslink-9.11.bin
          
"""

import sys,os,time,socket,signal,struct,zipfile

write = sys.stdout.write

if hasattr(signal,"alarm"):
    alarm = signal.alarm
else:
    def alarm(n):
        pass

def toHex(d):
    return ' '.join(['%02x' % ord(b) for b in d])

def fmtIP(d):
    return "%d.%d.%d.%d" % (((d>>24)&0xff),((d>>16)&0xff),((d>>8)&0xff),(d&0xff))

def toHex(d):
    return " ".join(["%02x" % ord(b) for b in d])

def macToStr(d):
    return ":".join(["%02x" % ord(b) for b in d])

def parseFront(fmt,data):
    size = struct.calcsize(fmt)
    if len(data) < size:
        sys.stderr.write("not enough data left for format '%s': [%d] %s\n" % (fmt, len(data), toHex(data)))
        raise ValueError
    rtuple =  struct.unpack(fmt, data[:size])
    rdata = data[size:]
    return rtuple,rdata

def decodeFeatureBits(f):
    names = [name for  (name,mask) in [("TelnetEnable",0x01), ("SnmpEnable",0x02), ("SecureDataEnable",0x04), ("SecureConfigEnable",0x08)] if (f&mask)]
    return "FeatureBits: [" + ",".join(names) + "]"


def reset(hostName):
    """Reset a devicemaster using TCP Admin command"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    alarm(10)
    s.connect((hostName,4606))
    s.send(struct.pack("<BBHL",13,0,8,0))
    s.close()
    alarm(0)

def getIdentReply(s):

    signal.alarm(10)
    cmd=5
    index=0
    length=8
    addr=0

    s.send(struct.pack("<BBHL",cmd,index,length,addr))
    resp = s.recv(1024)
    signal.alarm(0)

    if verbose>1:
        print "[%d] %s" % (len(resp),toHex(resp))

    return resp

def getArch(hostName):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    alarm(10)
    s.connect((hostName,4606))
    resp = getIdentReply(s)
    s.close()
    return ord(resp[98])

def isRedBoot(s,verbose=False):

    resp = getIdentReply(s)

    fmt = ">LLLLLL64s6sLBBHBBBB"

    devIpAddr, \
    defRoute, \
    srvIpAddr, \
    ipMask, \
    availMemAddr, \
    availMemLength, \
    version, \
    macAddr, \
    modelId, \
    archId, \
    numPorts, \
    progId, \
    boardRev, \
    bootRevMajor, \
    bootRevMinor, \
    numApps = struct.unpack(fmt,resp[:struct.calcsize(fmt)])

    remaining = resp[struct.calcsize(fmt):]

    appRecords = []
    extData = []
    appEnables = 0

    for i in range(numApps):
        appfmt = ">HBB"
        appRecords.append(struct.unpack(appfmt,remaining[:struct.calcsize(appfmt)]))
        remaining = remaining[struct.calcsize(appfmt):]

    (numUserBytes,),remaining = parseFront("B",remaining)

    if numUserBytes != 0:
        raise Exception("numUserBytes required to be 0")

    if remaining:
        (appEnables,),remaining = parseFront("B",remaining)
        while remaining:
            (etype,elen),remaining = parseFront("BB",remaining)
            edata = remaining[:elen]
            remaining = remaining[elen:]
            extData.append((etype,elen,edata))

    if verbose:
        print "%15s: %s" % ("devIpAddr",fmtIP(devIpAddr))
        print "%15s: %s" % ("defRoute",fmtIP(defRoute))
        print "%15s: %s" % ("srvIpAddr",fmtIP(srvIpAddr))
        print "%15s: %s" % ("ipMask",fmtIP(ipMask))
        print "%15s: %s" % ("availMemAddr",hex(availMemAddr))
        print "%15s: %s" % ("availMemLength",hex(availMemLength))
        if verbose>1:
            print "%15s: %s" % ("version",repr(version))
        else:
            print "%15s: %s" % ("version",version[:version.index('\x00')])  # ignore trailing NULs
        print "%15s: %s" % ("macAddr", macToStr(macAddr))
        print "%15s: %s" % ("modelId",modelId)
        print "%15s: %s" % ("archId",archId)
        print "%15s: %s" % ("numPorts",numPorts)
        print "%15s: %s" % ("progId",progId)
        print "%15s: %s" % ("boardRev",boardRev)
        print "%15s: %s" % ("bootRevMajor",bootRevMajor)
        print "%15s: %s" % ("bootRevMinor",bootRevMinor)
        print "%15s: %s" % ("numApps",numApps)
        print "%15s: %s" % ("appRecords",appRecords)
        print "%15s: %s" % ("numUserBytes",numUserBytes)
        print "%15s: %s" % ("appEnables",hex(appEnables))
        print "%15s:" % ("extData:")
        for e in extData:
            descr = "             type:%d len:%d data: %s" % (e[0],e[1],toHex(e[2]))
            if e[0] == 2:
                descr += " -- " + decodeFeatureBits(ord(e[2][0]))
            elif e[0] == 3:
                descr += " -- bootloader timeout"
            elif e[0] == 4:
                descr += " -- ipv6 mode"
            elif e[0] == 5:
                descr += " -- ipv4 configured address"
            elif e[0] == 6:
                descr += " -- ipv6 current address"
            print descr
    
    return version.startswith("Bootloader") and (progId == 0)


def downloadFile(hostName,fileobj,go=0,quiet=0,skip=0,base=0,entry=None,verify_redboot=0):
    """Download a binary file using TCP Admin command"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    alarm(10)
    s.connect((hostName,4606))

    isrb = isRedBoot(s,verbose)

    alarm(0)

    if verify_redboot and not isrb:
        print "Failed to re-connect to RedBoot, found running application."
        sys.exit(1)

    bytecount = len(fileobj.read(skip))
    index = 0
    while 1:
        alarm(10)
        data = fileobj.read(1400)
        if not data: break;
        length = len(data)
        dest = base+bytecount
        header = struct.pack("<BBHL",1,index,length+8,dest)
        if not quiet: write("%08x:%d\n" % (dest,length))
        time.sleep(0.005)
        s.send(header+data)
        bytecount = bytecount + length
        index += 1
        index &= 0xff
    if not quiet: write("downloaded %d bytes\n" % bytecount)
    if go:
        if not quiet: write("starting program\n")
        if entry is None:
            entry = base+skip
        s.send(struct.pack("<BBHL",3,0,8,entry))
    s.close()
    
args = sys.argv

progname = args[0]
del args[0]

e = None
g = 0
q = 0
r = 0
s = 0
b = 0
verbose = 0

def usage():
    write(usageString % progname)
    sys.exit(1)
    
while args and args[0].startswith('-'):
    if args[0][1] == 'b':
        if len(args[0]) > 2:
            b = int(args[0][2:],0)
        else:
            b = int(args[1],0)
            del args[1]
    elif args[0][1] == 'e':
        if len(args[0]) > 2:
            e = int(args[0][2:],0)
        else:
            e = int(args[1],0)
            del args[1]
    else:
        for opt in args[0][1:]:
            if   opt == 'g': g = 1
            elif opt == 's': s = 0x140
            elif opt == 'q': q = 1
            elif opt == 'r': r = 1
            elif opt == 'v': verbose += 1
            else: usage()
    del args[0]

if e is None:
    e = b

host = args[0]
filename = args[1]

if not zipfile.is_zipfile(filename):
    binfile = open(filename,"rb")
else:
    archsuffix = {3:".bin", 5:".nxbin"}
    write("%s is zipfile\n" % filename)
    arch = getArch(host)
    write("device arch == %d\n" % arch)
    zf = zipfile.ZipFile(filename,'r')
    for n in zf.namelist():
        if n.endswith(archsuffix[arch]):
            binfile = zf.open(n,'r')
            write("using binary '%s'\n" % n)
            break
    else:
        sys.stderr.write("ERROR: couldn't find binary in '%s' for arch %d\n" % (filename,arch))
        sys.exit(1)
    
if len(args)==1 and r:
    reset(host)
elif len(args)==2:    
    if r: 
        reset(host)
        time.sleep(3.0)
    downloadFile(host,binfile,go=g,quiet=q,skip=s,base=b,entry=e,verify_redboot=r)
else:
    usage()

sys.exit(0)
